@@ -129,6 +129,57 @@ sub chroot_tarball_if_too_old {
 	return undef;
 }
 
+sub get_extra_args {
+	my $self = shift;
+	# don't use $self->get_conf('DISTRIBUTION') as $dist might've been aliased
+	my $dist      = shift;
+	my $chroot    = shift;
+	my $arch      = $self->get_conf('BUILD_ARCH');
+	my $host_arch = $self->get_conf('HOST_ARCH');
+	if (!scalar(@{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') })) {
+		return [];
+	}
+	if (scalar(@{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') }) % 2 != 0)
+	{
+		print STDERR "W: length of UNSHARE_MMDEBSTRAP_EXTRA_ARGS is uneven\n";
+	}
+	my $extraargs = [];
+	my $it = natatime 2, @{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') };
+	while (my ($k, $v) = $it->()) {
+		if (ref($v) ne "ARRAY") {
+			print STDERR (
+					"W: entry $k in UNSHARE_MMDEBSTRAP_EXTRA_ARGS should be "
+				  . "an ARRAY but is a "
+				  . ref($v)
+				  . "\n");
+			next;
+		}
+		foreach
+		  my $arg ($dist, "$dist-$arch", "$dist-$arch-$host_arch", $chroot) {
+			next if !defined $arg;
+			if ($self->get_conf('DEBUG')) {
+				print STDERR "D: extra_args checking $arg\n";
+			}
+			if (ref($k) eq "Regexp") {
+				if ($arg !~ $k) {
+					next;
+				}
+			} elsif ($k ne $arg) {
+				next;
+			}
+			if ($self->get_conf('DEBUG')) {
+				print STDERR "D: extra_args $k matched\n";
+			}
+			print STDERR ("I: $k matched $arg -- adding extra arguments: "
+				  . (join " ", @{$v})
+				  . "\n");
+			push @{$extraargs}, @{$v};
+			last;
+		}
+	}
+	return $extraargs;
+}
+
 sub chroot_auto_create {
 	my $self      = shift;
 	my $chroot    = shift;
@@ -177,53 +228,17 @@ sub chroot_auto_create {
 		$dist = $newdist;
 	}
 
-	my @commonargs = (
-		$self->get_conf("MMDEBSTRAP"),
-		"--variant=buildd", "--arch=$arch", "--skip=output/mknod",
-		"--format=tar",     $basedist,
-	);
-	my $extraargs = [];
-	if (scalar(@{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') })) {
-		if (
-			scalar(@{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') }) % 2
-			!= 0) {
-			print STDERR
-			  "W: length of UNSHARE_MMDEBSTRAP_EXTRA_ARGS is uneven\n";
-		}
-		my $it = natatime 2,
-		  @{ $self->get_conf('UNSHARE_MMDEBSTRAP_EXTRA_ARGS') };
-		while (my ($k, $v) = $it->()) {
-			if (ref($v) ne "ARRAY") {
-				print STDERR (
-					"W: entry $k in UNSHARE_MMDEBSTRAP_EXTRA_ARGS should be "
-					  . "an ARRAY but is a "
-					  . ref($v)
-					  . "\n");
-				next;
-			}
-			foreach
-			  my $arg ($dist, "$dist-$arch", "$dist-$arch-$host_arch", $chroot)
-			{
-				next if !defined $arg;
-				if ($self->get_conf('DEBUG')) {
-					print STDERR "D: extra_args checking $arg\n";
-				}
-				if (ref($k) eq "Regexp") {
-					if ($arg !~ $k) {
-						next;
-					}
-				} elsif ($k ne $arg) {
-					next;
-				}
-				if ($self->get_conf('DEBUG')) {
-					print STDERR "D: extra_args $k matched\n";
-				}
-				push @{$extraargs}, @{$v};
-				last;
-			}
-		}
+	my @commonargs = ();
+	if (length $self->get_conf('UNSHARE_MMDEBSTRAP_ENV_CMD')) {
+		push(@commonargs, @{ $self->get_conf('UNSHARE_MMDEBSTRAP_ENV_CMD') });
 	}
-
+	push(
+		@commonargs,
+		(
+			$self->get_conf("MMDEBSTRAP"), "--variant=buildd",
+			"--arch=$arch",                "--skip=output/mknod",
+			"--format=tar",                $basedist,
+		));
 	if ($self->get_conf('UNSHARE_MMDEBSTRAP_KEEP_TARBALL')) {
 		# the tarball is supposed to be kept but maybe we don't need to
 		# create one because the existing one is new enough
@@ -246,6 +261,8 @@ sub chroot_auto_create {
 			$chroot  = "$xdg_cache_home/$dist-$arch.tar";
 			$self->set('Chroot ID', $chroot);
 		}
+		my $extraargs = $self->get_extra_args($dist, $chroot);
+
 		print STDERR ("I: Creating new chroot tarball:\n"
 			  . join(" ", (@commonargs, $tarball, @{$extraargs}))
 			  . "\n");
@@ -273,6 +290,8 @@ sub chroot_auto_create {
 		return $chroot;
 	}
 
+	my $extraargs = $self->get_extra_args($dist, $chroot);
+
 	# chroot was found but UNSHARE_MMDEBSTRAP_KEEP_TARBALL was
 	# false so if the existing tarball is too old, don't use it
 	# if Chroot ID was undefined, then Sbuild::ChrootInfoUnshare
@@ -724,6 +743,23 @@ sub begin_session {
 
 	$self->set('Session ID', $rootdir);
 
+	# Run /bin/true inside the chroot to prevent really bad error messages
+	# later on. If sbuild-usernsexec fails due to a bug or bad setup, then
+	# later setup code which relies on the exit status of a command that
+	# is run inside the chroot will end up doing the wrong thing with
+	# really misleading error messages. See this MR and related commit:
+	# https://salsa.debian.org/debian/sbuild/-/merge_requests/173
+	$self->run_command({
+		COMMAND => ['true'],
+		USER    => 'root',
+		DIR     => '/'
+	});
+	if ($?) {
+		print STDERR ("E: Running 'true' inside the chroot failed. "
+			  . "Please file a bug against sbuild.\n");
+		return 0;
+	}
+
 	$self->set('Location', '/sbuild-unshare-dummy-location');
 
 	$self->set('Session Purged', 1);
@@ -774,13 +810,27 @@ sub end_session {
 	  . $self->get('Session ID') . ")\n"
 	  if $self->get_conf('DEBUG');
 
-	# this looks like a recipe for disaster, but since we execute "rm -rf" with
-	# lxc-usernsexec, we only have permission to delete the files that were
-	# created with the fake root user
+	# This looks like a recipe for disaster, but since we execute "find
+	# -delete" with lxc-usernsexec, we only have permission to delete the
+	# files that were created with the fake root user that is different to the
+	# user executing sbuild.
+	# We change CWD to the chroot directory because find tries to chdir to the
+	# current directory which might not be accessible by the unshared user:
+	# find: Failed to restore initial working directory
+	# We don't use rm -rf as that could print "Permission denied".
 	my @cmd = (
 		"/usr/libexec/sbuild-usernsexec",
 		(map { join ":", @{$_} } @idmap),
-		'--', 'rm', '-rf', $self->get('Session ID'));
+		'--',
+		'env',
+		("--chdir=" . $self->get('Session ID')),
+		'find',
+		$self->get('Session ID'),
+		'-mount',
+		'-mindepth',
+		'1',
+		'-delete'
+	);
 	if ($self->get_conf('DEBUG')) {
 		printf STDERR "running @cmd\n";
 	}
@@ -788,6 +838,22 @@ sub end_session {
 	# we ignore the exit status, because the command will fail to remove the
 	# unpack directory itself because of insufficient permissions
 
+	# After removing the chroot directory content above we also need to
+	# remove the directory itself. Use unshare with uid 0 here, same as how
+	# it was created begin_session.
+	@cmd = (
+		'unshare',
+		# comment to guide perltidy line wrapping
+		'--map-user',   '0',
+		'--map-group',  '0',
+		'--map-users',  "$idmap[0][2],1,1",
+		'--map-groups', "$idmap[1][2],1,1",
+		'rmdir',        '--', $self->get('Session ID'));
+	if ($self->get_conf('DEBUG')) {
+		printf STDERR "running @cmd\n";
+	}
+	system(@cmd);
+
 	if (-d $self->get('Session ID') && !rmdir($self->get('Session ID'))) {
 		print STDERR "unable to remove " . $self->get('Session ID') . ": $!\n";
 		$self->set('Session ID', "");
