@@ -105,13 +105,6 @@ sub set_options {
 			if ($opt_no_arch_all) {
 				die "--arch-all cannot be used together with --no-arch-all";
 			}
-			if ($opt_make_binnmu) {
-				die "--arch-all cannot be used together with --make-binNMU";
-			}
-			if ($opt_append_to_version) {
-				die
-"--arch-all cannot be used together with --append-to-version";
-			}
 			$self->set_conf('BUILD_ARCH_ALL', 1);
 			$opt_arch_all = 1;
 		},
@@ -351,6 +344,12 @@ sub set_options {
 			$self->set_conf('MAILTO',               $_[1]);
 			$self->set_conf('MAILTO_FORCED_BY_CLI', "yes");
 		},
+		"hide-log-sections=s" => sub {
+			$self->push_conf('LOG_HIDDEN_SECTIONS', split(/,/, $_[1]));
+		},
+		"log-filename-ts-format=s" => sub {
+			$self->set_conf('LOG_FILENAME_TIMESTAMP_FORMAT', $_[1]);
+		},
 		"n|nolog" => sub {
 			$self->set_conf('NOLOG', 1);
 		},
@@ -694,6 +693,9 @@ sub set_options {
 			$self->push_conf('EXTRA_REPOSITORY_KEYS', $_[1]);
 		},
 		"build-path=s" => sub {
+			if ($_[1] !~ /^($|\/)/) {
+				die "--build-path must be an absolute path";
+			}
 			$self->set_conf('BUILD_PATH', $_[1]);
 		},
 		"dsc-dir=s" => sub {
